/*
 * Copyright (c) 2003 Linuxant inc.
 *
 * NOTE: The use and distribution of this software is governed by the terms in
 * the file LICENSE, which is included in the package. You must read this and
 * agree to these terms before using or distributing this software.
 *
 */

#define OS_NOSTRINGREDEFS

#include "oscompat.h"
#include "osservices.h"
#include "ndiswrap.h"
#include "netndis.h"

#include <linux/pci.h>
#include <linux/kmod.h>
#include <linux/ctype.h>

static struct pci_device_id dldrpci_tbl[] __devinitdata =
{
	{PCI_ANY_ID, PCI_ANY_ID, PCI_ANY_ID, PCI_ANY_ID, 0, 0, 0},
	{0,}                    /* terminate list */
};


MODULE_DEVICE_TABLE(pci, dldrpci_tbl);

MODULE_AUTHOR("Copyright (C) 2003 Linuxant inc.");
MODULE_DESCRIPTION(DLDRDESC);
MODULE_LICENSE("see LICENSE file; Copyright (c)2003 Linuxant inc.");

static int dldrpci_suspend(struct pci_dev *pdev, u32 state)
{
	POS_DEVNODE pDevNode = pci_get_drvdata(pdev);
	struct net_device *ndev = pDevNode->netDev;
	struct netndis_private *priv = ndev->priv;
	int err;
	
	printk(KERN_DEBUG "%s: entering sleep mode (state=%d)\n",
			ndev->name, state);

	err = netndis_lock(priv, FALSE);
	if (err) {
		printk(KERN_ERR "%s: hw_unavailable on %s\n", ndev->name, __FUNCTION__);
		return err;
	}

	err = __netndis_down(ndev);
	if (err)
		printk(KERN_WARNING "%s: %s: Error %d downing interface\n", ndev->name, __FUNCTION__, err);
	
	netif_device_detach(ndev);

	priv->hw_unavailable++;

#if 0
	pci_save_state(pdev, pDevNode->hwPmState);
	pci_disable_device(pdev);
	pci_set_power_state(pdev, 3);
#endif

	netndis_unlock(priv);

	return 0;
}

static int dldrpci_resume(struct pci_dev *pdev)
{
	POS_DEVNODE pDevNode = pci_get_drvdata(pdev);
	struct net_device *ndev = pDevNode->netDev;
	struct netndis_private *priv = ndev->priv;
	int err;

	printk(KERN_DEBUG "%s: waking up\n", ndev->name);

#if 0
	pci_enable_device(pdev);
	pci_restore_state(pdev, pDevNode->hwPmState);

	err = netndis_resume(ndev);
	if (err) {
		printk(KERN_ERR "%s: Error %d on netndis_resume()\n", ndev->name, err);
		return err;
	}
#endif

	err = netndis_lock(priv, TRUE);
	if (err) {
		printk(KERN_ERR "%s: hw_unavailable on %s\n", ndev->name, __FUNCTION__);
		return err;
	}

	netif_device_attach(ndev);

	priv->hw_unavailable--;

	if (priv->open && (! priv->hw_unavailable)) {
		err = __netndis_up(ndev);
		if (err)
			printk(KERN_WARNING "%s: %s: Error %d restarting interface\n", ndev->name, __FUNCTION__, err);
	}
	
	netndis_unlock(priv);

	return 0;
}

static void __devexit dldrpci_remove (struct pci_dev *pdev)
{
	POS_DEVNODE pDevNode = pci_get_drvdata(pdev);

	if(pDevNode->netDev) {
		unregister_netdev(pDevNode->netDev);
	}

	NdisWrapHaltAdapter(pDevNode->miniportAdapter);

	pci_set_drvdata(pdev, NULL);

	free_netdev(pDevNode->netDev);
	NdisWrapDestroyAdapter(pDevNode->miniportAdapter);
	kfree(pDevNode);

	pci_disable_device(pdev);
	pci_release_regions(pdev);
}


static int __devinit dldrpci_probe (struct pci_dev *pdev,
		const struct pci_device_id *ent)
{
	int err;
	POS_DEVNODE pDevNode;
	struct net_device *ndev = NULL;
	char buf[100];
    char *argv[3], *envp[3], *p;
	u_int8_t csz;

	pDevNode = kmalloc(sizeof(*pDevNode), GFP_KERNEL);
	if(!pDevNode) {
		return -ENOMEM;
	}

	memset(pDevNode, 0, sizeof(*pDevNode));
	pDevNode->hwDev = pdev;
	pDevNode->hwModule = THIS_MODULE;
	pDevNode->hwBusType = 5; /* PCI */

	/* count : chars in slot_name to determine if fake domain_nr must be added to slot_name
	 * to make 2.4 instance names identical to 2.6 */
	for(p = pdev->slot_name, csz = 0; *p; p++)
		if(*p == ':')
			csz++;
	snprintf(pDevNode->hwInstName, sizeof(pDevNode->hwInstName), "PCI-%s%s-%04x:%04x-%04x:%04x", csz == 1 ? "0000:" : "", pdev->slot_name, pdev->vendor, pdev->device, pdev->subsystem_vendor, pdev->subsystem_device);

	err = NdisWrapCreateAdapter(pDevNode, DLDRVARDIR, &pDevNode->miniportAdapter);
	OsDebugPrintf("%s: NdisWrapCreateAdapter returned %d miniportAdapter=%p\n", __FUNCTION__, err, pDevNode->miniportAdapter);
	if(err) {
		kfree(pDevNode);
		return err;
	}

	if (pci_request_regions(pdev, (char*)(THIS_MODULE->name))) {
		printk(KERN_ERR "%s: pci_request_regions(%s) failed\n", pci_name(pdev), pdev->slot_name); 
		NdisWrapDestroyAdapter(pDevNode->miniportAdapter);
		kfree(pDevNode);
		return -ENODEV;
	}

	if ( (err = pci_enable_device(pdev)) ) {
		printk(KERN_ERR "%s: pci_enable_device(%s) failed (err=%d)\n", pci_name(pdev), pdev->slot_name, err);
		pci_release_regions(pdev);
		NdisWrapDestroyAdapter(pDevNode->miniportAdapter);
		kfree(pDevNode);
		return err;
	}

	/* XXX 32-bit addressing only */
	if (pci_set_dma_mask(pdev, 0xffffffff)) {
		printk(KERN_ERR "%s: 32-bit DMA not available\n", pci_name(pdev));
		pci_release_regions(pdev);
		NdisWrapDestroyAdapter(pDevNode->miniportAdapter);
		kfree(pDevNode);
		return -EIO;
	}

	/*
	 * Cache line size is used to size and align various
	 * structures used to communicate with the hardware.
	 */
	pci_read_config_byte(pdev, PCI_CACHE_LINE_SIZE, &csz);
	if (csz == 0) {
		/*
		 * Linux 2.4.18 (at least) writes the cache line size
		 * register as a 16-bit wide register which is wrong.
		 * We must have this setup properly for rx buffer
		 * DMA to work so force a reasonable value here if it
		 * comes up zero.
		 */
		csz = L1_CACHE_BYTES / sizeof(u_int32_t);
		printk("%s: cache line size not set; forcing %u\n", pci_name(pdev), csz);
		pci_write_config_byte(pdev, PCI_CACHE_LINE_SIZE, csz);
	}
	/*
	 * The default setting of latency timer yields poor results,
	 * set it to the value used by other systems.  It may be worth
	 * tweaking this setting more.
	 */
	pci_write_config_byte(pdev, PCI_LATENCY_TIMER, 0xa8);

	pci_set_drvdata(pdev, pDevNode);

	ndev = alloc_netndisdev(pDevNode);
	if(!ndev) {
		pci_set_drvdata(pdev, NULL);
		pci_disable_device(pdev);
		pci_release_regions(pdev);
		NdisWrapDestroyAdapter(pDevNode->miniportAdapter);
		kfree(pDevNode);
		return -ENOMEM;
	}

	SET_MODULE_OWNER(ndev);

	pDevNode->netDev = ndev;

	err = register_netdev(ndev);
	if (err) {
		printk(KERN_ERR "%s: Failed to register net device (%d)\n", ndev->name, err);
		pDevNode->netDev = NULL;
		kfree(ndev);
		pci_set_drvdata(pdev, NULL);
		pci_disable_device(pdev);
		pci_release_regions(pdev);
		NdisWrapDestroyAdapter(pDevNode->miniportAdapter);
		kfree(pDevNode);
		return err;
	}

	buf[0] = '\0';
	NdisWrapGetVendorDesc(pDevNode->miniportAdapter, buf, sizeof(buf));

	printk(KERN_DEBUG"%s: %s at %s (MAC address %02X:%02X:%02X:%02X:%02X:%02X) ready\n", ndev->name, buf,
			pci_name(pdev), 
			ndev->dev_addr[0], ndev->dev_addr[1], ndev->dev_addr[2],
			ndev->dev_addr[3], ndev->dev_addr[4], ndev->dev_addr[5]);

    envp [0] = "HOME=/";
    envp [1] = "PATH=/sbin:/bin:/usr/sbin:/usr/bin:"DLDRSBINDIR;
    envp [2] = 0;

    argv[0] = DLDRSBINDIR"/dldrconfig";
    argv[1] = "--auto";
    sprintf(buf, "--netif=%s", ndev->name);
    argv[2] = buf;
    argv[2] = "--modalias";
    argv[3] = 0;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
    err = call_usermodehelper(argv[0], argv, envp);
#else
    err = call_usermodehelper(argv[0], argv, envp, 1);
#endif
    if(err) {
        printk(KERN_ERR"%s: call_usermodehelper %s returned %d\n", __FUNCTION__, argv[0], err);
    }

	return 0;
}

static struct pci_driver dldrpci_driver = {
	.name           = "driverloader",
	.id_table       = dldrpci_tbl,
	.probe          = dldrpci_probe,
	.remove         = __devexit_p(dldrpci_remove),

	.suspend        = dldrpci_suspend,
	.resume         = dldrpci_resume,
};


static int __init dldrpci_init (void)
{
	int err;
    char *argv[2], *envp[4];

    envp [0] = "HOME=/";
    envp [1] = "PATH=/sbin:/bin:/usr/sbin:/usr/bin:"DLDRSBINDIR;
    envp [2] = "DLDRHELPERACTION=start";
    envp [3] = 0;
    argv[0] = DLDRLIBDIR"/webconfd";
    argv[1] = 0;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
    err = call_usermodehelper(argv[0], argv, envp);
#else
    err = call_usermodehelper(argv[0], argv, envp, 1);
#endif
    if(err) {
        printk(KERN_ERR"%s: call_usermodehelper %s returned %d\n", __FUNCTION__, argv[0], err);
    }

	netndis_init();
	err = NdisWrapInit();
	if(err) {
		netndis_exit();
		return err;
	}
	err = pci_module_init (&dldrpci_driver);
	if(err) {
		NdisWrapExit();
		netndis_exit();
	}
	return err;
}


static void __exit dldrpci_cleanup (void)
{
	pci_unregister_driver (&dldrpci_driver);
	NdisWrapExit();
	netndis_exit();
}

module_init(dldrpci_init);
module_exit(dldrpci_cleanup);
