/*
 * Copyright (c) 2003 Linuxant inc.
 *
 * NOTE: The use and distribution of this software is governed by the terms in
 * the file LICENSE, which is included in the package. You must read this and
 * agree to these terms before using or distributing this software.
 *
 */
#include "oscompat.h"
#include "typedefs.h"
#include "osservices.h"

#include <linux/slab.h>
#include <linux/vmalloc.h>
#include <linux/highmem.h>
#include <linux/pci.h>

__shimcall__
void *OsAllocate(unsigned size)
{
	PVOID pMem = kmalloc(size, OsContextAllowsSleeping() ? GFP_KERNEL : GFP_ATOMIC);
	if(!pMem) {
		if(OsContextAllowsSleeping()) {
			/* try vmalloc since kmalloc is generally limited to 64k */
			pMem = vmalloc(size);
			//printk(KERN_DEBUG "%s(%u): vmalloc returned %p\n", __FUNCTION__, size, pMem);
		}
	}

	if(pMem) {
#ifdef OS_DEBUG_MEM
		printk(KERN_DEBUG "%s(%u): %p\n", __FUNCTION__, size, pMem);
#endif
		memset(pMem, 0, size); /* caller expects memory to be initialized to zero */
	} else {
		printk(KERN_ERR "%s(%u): failed, OsContextAllowsSleeping=%s\n", __FUNCTION__, size, OsContextAllowsSleeping() ? "TRUE" : "FALSE");
	}

	return pMem;
}
 
__shimcall__
void OsFree(PVOID pMem)
{
	if(!pMem) {
		printk(KERN_ERR "%s: NULL\n", __FUNCTION__);
		return;
	}
	if((pMem >= (PVOID)VMALLOC_START) && (pMem < (PVOID)VMALLOC_END)) {
#ifdef OS_DEBUG_MEM
		printk(KERN_DEBUG "%s: vfree %p\n", __FUNCTION__, pMem);
#endif
		if(!(OsContextAllowsSleeping())) {
			printk(KERN_WARNING "%s: calling vfree(%p) in non-sleeping context\n", __FUNCTION__, pMem);
		}
		vfree(pMem);
	} else {
#ifdef OS_DEBUG_MEM
		printk(KERN_DEBUG "%s: %p\n", __FUNCTION__, pMem);
#endif
		kfree(pMem);
	}
}

__shimcall__
void *OsDmaAllocate(unsigned size, unsigned long *physAddr)
{
	return pci_alloc_consistent(NULL, size, (dma_addr_t*)physAddr);
}

__shimcall__
void OsDmaFree(unsigned size, void *vAddr, unsigned long physAddr)
{
	pci_free_consistent(NULL, size, vAddr, physAddr);
}
